from typing import Any, Optional, Dict
from datetime import datetime, timezone, timedelta
from .CacheItem import CacheItem


class Cache:

    def __init__(self, lru_max_size: int = None):
        self._lru_max_size = lru_max_size
        self._items = {}


    def get(self, key: str) -> Optional[Any]:
        if key not in self._items:
            return

        item = self._items[key]

        if self._is_expired(item):
            return

        return item.value


    def set(self, key: str, value: Any, ttl_in_seconds: int = None):
        if value is None:
            raise RuntimeError(f'Attempted to cache None for key: {key}')

        expires_at = None
        if ttl_in_seconds is not None:
            expires_at = datetime.now(timezone.utc) + timedelta(seconds=ttl_in_seconds)

        self._items[key] = CacheItem(value, expires_at)
        self._cleanup()


    def get_all(self) -> Dict[str, Any]:
        values = {}
        for key in self._items:
            value = self.get(key)
            if value is not None:
                values[key] = value

        return values


    def delete(self, key: str):
        del self._items[key]


    def has(self, key: str) -> bool:
        return self.get(key) is not None


    def _cleanup(self):
        # delete oldest
        while len(self._items) > self._lru_max_size:
            self._items.popitem()

        # delete expired
        for key, item in list(self._items.items()):
            if self._is_expired(item):
                self.delete(key)


    def _is_expired(self, item: CacheItem) -> bool:
        return item.expires_at and item.expires_at < datetime.now(timezone.utc)


    # """ Make it iterable """

    # def __len__(self):
    #     return len(self._items)


    # def __getitem__(self, index):
    #     values = list(self._items.values())
    #     return values[index]
