from __future__ import annotations
from typing import List
import inspect
from sanic.log import logger
from ..caching.Cache import Cache
from .ClassParam import ClassParam
from .ContextualService import ContextualService


class DiContainer:

    MAX_CACHE_SIZE = 100


    def __init__(self):
        self._services = Cache(self.MAX_CACHE_SIZE)
        key = self._get_key(type(self))
        self._services.set(key, self)


    def get(self, cls: type, **kwargs) -> object:
        key = self._get_key(cls)

        if self._services.has(key):
            return self._services.get(key)

        instance = self.create(cls, **kwargs)
        self._set(instance)
        return instance


    def init(self, cls: type, **kwargs):
        key = self._get_key(cls)

        if self._services.has(key):
            logger.debug(f'DiContainer: Init failed - {cls} already instantiated')

        self.get(cls, **kwargs)


    def create(self, cls: type, **kwargs) -> object:
        logger.debug(f'DiContainer: Instantiating {cls}')

        inject = {}
        for param in self._get_class_params(cls):
            if param.name in kwargs:
                inject[param.name] = kwargs[param.name]
                continue

            if param.has_default():
                continue

            inject[param.name] = self.get(param.type)

        return cls(**inject)


    def fork(self) -> DiContainer:
        logger.debug(f'DiContainer: Forking')

        container = DiContainer()

        for service in self._services.get_all().values():
            container._set(service)

        return container


    def teardown(self):
        logger.debug(f'DiContainer: Teardown')

        for key, service in self._services.get_all().copy().items():
            if isinstance(service, ContextualService):
                service.on_context_teardown()
                self._services.delete(key)


    def _set(self, instance: object):
        cls = type(instance)
        key = self._get_key(cls)
        self._services.set(key, instance)


    def _get_class_params(self, cls: type) -> List[ClassParam]:
        signature = inspect.signature(cls.__init__)

        params = []
        for key, param in signature.parameters.items():
            if param.name in ('self', 'args', 'kwargs'):
                continue

            params.append(ClassParam(param.name, param.annotation, param.default))

        return params


    @staticmethod
    def _get_key(cls: type) -> str:
        return cls.__qualname__
