from typing import Callable
from hashlib import sha1
from sanic import Sanic, response
from sanic.log import logger
from jinja2 import Environment, FileSystemLoader, select_autoescape
from functools import wraps
from ..config.ConfigProvider import ConfigProvider
from ..caching.Cache import Cache
from .ContextualViewDecorator import ContextualViewDecorator


class JinjaRenderer(ContextualViewDecorator):

    def __init__(self, app: Sanic, config_provider: ConfigProvider):
        self._app = app
        self._config = config_provider.get_section('html')
        self._environment = Environment(
            # loader=PackageLoader(__name__, templates_dir),
            loader=FileSystemLoader(self._config['templatesDir']),
            autoescape=select_autoescape(),
            enable_async=True
        )
        self._environment.filters['url'] = self._get_url_filter()
        self._cache = Cache(self._config['cache']['maxSize'])


    def get_decorator(self) -> Callable:
        def apply_template(func):
            @wraps(func)
            async def wrapper(*args, **kwargs) -> str:
                request = args[0]
                view_name = request.endpoint.split('.')[-1]
                template_name = view_name + '.html'

                # try cache
                cache_key = self._cache_key(template_name, **kwargs)
                html = self._cache.get(cache_key)
                if html:
                    logger.debug(f'JinjaRenderer: Using cache for {template_name}')
                    return response.html(html)

                # render
                view_result = await func(*args, **kwargs)

                # if view returned response, return it
                if isinstance(view_result, response.BaseHTTPResponse):
                    return view_result

                template = self._environment.get_template(template_name)
                html = await template.render_async(view_result)

                # set cache
                self._cache.set(cache_key, html, self._config['cache']['ttlInSeconds'])

                return response.html(html)

            return wrapper
        return apply_template


    def _get_url_filter(self) -> Callable:
        def url_filter(view_name: str, **kwargs) -> str:
            return self._app.url_for(view_name, **kwargs)
        return url_filter


    @staticmethod
    def _cache_key(template_name, **kwargs) -> str:
        tup = (template_name,) + tuple(sorted(kwargs.items()))  # https://stackoverflow.com/a/10220908
        key = ''.join(map(str, tup))
        hash = sha1(key.encode('utf-8')).hexdigest()
        return hash
